@@ -452,7 +452,6 @@ meta_frames_calc_geometry (MetaFrames        *frames,
 MetaFrames*
 meta_frames_new (int screen_number)
 {
-#ifdef HAVE_GTK_MULTIHEAD
   GdkScreen *screen;
 
   screen = gdk_display_get_screen (gdk_display_get_default (),
@@ -460,12 +459,7 @@ meta_frames_new (int screen_number)
 
   return g_object_new (META_TYPE_FRAMES,
 		       "screen", screen,
-		       NULL);
-#else
-  return g_object_new (META_TYPE_FRAMES,
-		       NULL);
-#endif
-  
+		       NULL);  
 }
 
 void
@@ -495,6 +489,7 @@ meta_frames_manage_window (MetaFrames *frames,
   frame->text_height = -1;
   frame->title = NULL;
   frame->expose_delayed = FALSE;
+  frame->shape_applied = FALSE;
   frame->prelit_control = META_FRAME_CONTROL_NONE;
   
   meta_core_grab_buttons (gdk_display, frame->xwindow);
@@ -664,7 +659,8 @@ void
 meta_frames_apply_shapes (MetaFrames *frames,
                           Window      xwindow,
                           int         new_window_width,
-                          int         new_window_height)
+                          int         new_window_height,
+                          gboolean    window_has_shape)
 {
 #ifdef HAVE_SHAPE
   /* Apply shapes as if window had new_window_width, new_window_height */
@@ -685,10 +681,25 @@ meta_frames_apply_shapes (MetaFrames *frames,
   if (!(fgeom.top_left_corner_rounded ||
         fgeom.top_right_corner_rounded ||
         fgeom.bottom_left_corner_rounded ||
-        fgeom.bottom_right_corner_rounded))
+        fgeom.bottom_right_corner_rounded ||
+        window_has_shape))
     {
-      XShapeCombineMask (gdk_display, frame->xwindow,
-                         ShapeBounding, 0, 0, None, ShapeSet);
+      if (frame->shape_applied)
+        {
+          meta_topic (META_DEBUG_SHAPES,
+                      "Unsetting shape mask on frame 0x%lx\n",
+                      frame->xwindow);
+          
+          XShapeCombineMask (gdk_display, frame->xwindow,
+                             ShapeBounding, 0, 0, None, ShapeSet);
+          frame->shape_applied = FALSE;
+        }
+      else
+        {
+          meta_topic (META_DEBUG_SHAPES,
+                      "Frame 0x%lx still doesn't need a shape mask\n",
+                      frame->xwindow);
+        }
       
       return; /* nothing to do */
     }
@@ -804,12 +815,99 @@ meta_frames_apply_shapes (MetaFrames *frames,
 
   XSubtractRegion (window_xregion, corners_xregion, window_xregion);
 
-  XShapeCombineRegion (gdk_display, frame->xwindow,
-                       ShapeBounding, 0, 0, window_xregion, ShapeSet);
+  XDestroyRegion (corners_xregion);
+  
+  if (window_has_shape)
+    {
+      /* The client window is oclock or something and has a shape
+       * mask. To avoid a round trip to get its shape region, we
+       * create a fake window that's never mapped, build up our shape
+       * on that, then combine. Wasting the window is assumed cheaper
+       * than a round trip, but who really knows for sure.
+       */
+      XSetWindowAttributes attrs;      
+      Window shape_window;
+      Window client_window;
+      Region client_xregion;
+      GdkScreen *screen;
+      int screen_number;
+      
+      meta_topic (META_DEBUG_SHAPES,
+                  "Frame 0x%lx needs to incorporate client shape\n",
+                  frame->xwindow);
+
+      screen = gtk_widget_get_screen (GTK_WIDGET (frames));
+      screen_number = gdk_x11_screen_get_screen_number (screen);
+      
+      attrs.override_redirect = True;
+      
+      shape_window = XCreateWindow (gdk_display,
+                                    RootWindow (gdk_display, screen_number),
+                                    -5000, -5000,
+                                    new_window_width,
+                                    new_window_height,
+                                    0,
+                                    CopyFromParent,
+                                    CopyFromParent,
+                                    CopyFromParent,
+                                    CWOverrideRedirect,
+                                    &attrs);
+
+      /* Copy the client's shape to the temporary shape_window */
+      client_window = meta_core_get_client_xwindow (gdk_display,
+                                                    frame->xwindow);
+
+      XShapeCombineShape (gdk_display, shape_window, ShapeBounding,
+                          fgeom.left_width,
+                          fgeom.top_height,
+                          client_window,
+                          ShapeBounding,
+                          ShapeSet);
+
+      /* Punch the client area out of the normal frame shape,
+       * then union it with the shape_window's existing shape
+       */
+      client_xregion = XCreateRegion ();
+  
+      xrect.x = fgeom.left_width;
+      xrect.y = fgeom.top_height;
+      xrect.width = new_window_width - fgeom.right_width - xrect.x;
+      xrect.height = new_window_height - fgeom.bottom_height - xrect.y;
+
+      XUnionRectWithRegion (&xrect, client_xregion, client_xregion);
+      
+      XSubtractRegion (window_xregion, client_xregion, window_xregion);
+
+      XDestroyRegion (client_xregion);
+      
+      XShapeCombineRegion (gdk_display, shape_window,
+                           ShapeBounding, 0, 0, window_xregion, ShapeUnion);
+      
+      /* Now copy shape_window shape to the real frame */
+      XShapeCombineShape (gdk_display, frame->xwindow, ShapeBounding,
+                          0, 0,
+                          shape_window,
+                          ShapeBounding,
+                          ShapeSet);
+
+      XDestroyWindow (gdk_display, shape_window);
+    }
+  else
+    {
+      /* No shape on the client, so just do simple stuff */
+
+      meta_topic (META_DEBUG_SHAPES,
+                  "Frame 0x%lx has shaped corners\n",
+                  frame->xwindow);
+      
+      XShapeCombineRegion (gdk_display, frame->xwindow,
+                           ShapeBounding, 0, 0, window_xregion, ShapeSet);
+    }
+  
+  frame->shape_applied = TRUE;
   
   XDestroyRegion (window_xregion);
-  XDestroyRegion (corners_xregion);
-#endif
+#endif /* HAVE_SHAPE */
 }
 
 void
@@ -947,11 +1045,8 @@ show_tip_now (MetaFrames *frames)
       /* get conversion delta for root-to-frame coords */
       dx = root_x - x;
       dy = root_y - y;
-#ifdef HAVE_GTK_MULTIHEAD
       screen_number = gdk_screen_get_number (gtk_widget_get_screen (GTK_WIDGET (frames)));
-#else
-      screen_number = DefaultScreen (gdk_display);
-#endif
+      
       meta_fixed_tip_show (gdk_display,
 			   screen_number,
                            rect->x + dx,
@@ -1230,8 +1325,7 @@ meta_frames_button_press_event (GtkWidget      *widget,
                                event->x_root,
                                event->y_root);
     }
-  else if ((control == META_FRAME_CONTROL_TITLE ||
-            control == META_FRAME_CONTROL_NONE) &&
+  else if (control == META_FRAME_CONTROL_TITLE &&
            event->button == 1)
     {
       MetaFrameFlags flags;
@@ -1856,14 +1950,17 @@ get_control (MetaFrames *frames,
   MetaFrameFlags flags;
   gboolean has_vert, has_horiz;
   GdkRectangle client;
+  int bottom_of_titlebar;
   
   meta_frames_calc_geometry (frames, frame, &fgeom);
 
   client.x = fgeom.left_width;
   client.y = fgeom.top_height;
   client.width = fgeom.width - fgeom.left_width - fgeom.right_width;
-  client.height = fgeom.height - fgeom.top_height - fgeom.bottom_height;
+  client.height = fgeom.height - fgeom.top_height - fgeom.bottom_height;  
 
+  bottom_of_titlebar = fgeom.title_rect.y + fgeom.title_rect.height;
+  
   if (POINT_IN_RECT (x, y, client))
     return META_FRAME_CONTROL_CLIENT_AREA;
   
@@ -1875,16 +1972,19 @@ get_control (MetaFrames *frames,
 
   if (POINT_IN_RECT (x, y, fgeom.menu_rect))
     return META_FRAME_CONTROL_MENU;
+
+  flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
+  
+  has_vert = (flags & META_FRAME_ALLOWS_VERTICAL_RESIZE) != 0;
+  has_horiz = (flags & META_FRAME_ALLOWS_HORIZONTAL_RESIZE) != 0;
   
   if (POINT_IN_RECT (x, y, fgeom.title_rect))
     {
-      if (y <= TOP_RESIZE_HEIGHT)
+      if (has_vert && y <= TOP_RESIZE_HEIGHT)
         return META_FRAME_CONTROL_RESIZE_N;
       else
         return META_FRAME_CONTROL_TITLE;
     }
-  
-  flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
 
   if (POINT_IN_RECT (x, y, fgeom.max_rect))
     {
@@ -1893,83 +1993,78 @@ get_control (MetaFrames *frames,
       else
         return META_FRAME_CONTROL_MAXIMIZE;
     }
-  
-  has_vert = (flags & META_FRAME_ALLOWS_VERTICAL_RESIZE) != 0;
-  has_horiz = (flags & META_FRAME_ALLOWS_HORIZONTAL_RESIZE) != 0;
-
-  if (has_vert || has_horiz)
-    {
-      int bottom_of_titlebar;
-
-      bottom_of_titlebar = fgeom.title_rect.y + fgeom.title_rect.height;
       
-      /* South resize always has priority over north resize,
-       * in case of overlap.
-       */
+  /* South resize always has priority over north resize,
+   * in case of overlap.
+   */
 
-      if (y >= (fgeom.height - fgeom.bottom_height - RESIZE_EXTENDS) &&
-          x >= (fgeom.width - fgeom.right_width - RESIZE_EXTENDS))
-        {
-          if (has_vert && has_horiz)
-            return META_FRAME_CONTROL_RESIZE_SE;
-          else if (has_vert)
-            return META_FRAME_CONTROL_RESIZE_S;
-          else
-            return META_FRAME_CONTROL_RESIZE_E;
-        }
-      else if (y >= (fgeom.height - fgeom.bottom_height - RESIZE_EXTENDS) &&
-               x <= (fgeom.left_width + RESIZE_EXTENDS))
-        {
-          if (has_vert && has_horiz)
-            return META_FRAME_CONTROL_RESIZE_SW;
-          else if (has_vert)
-            return META_FRAME_CONTROL_RESIZE_S;
-          else
-            return META_FRAME_CONTROL_RESIZE_W;
-        }
-      else if (y < (fgeom.top_height + RESIZE_EXTENDS) &&
-               x < RESIZE_EXTENDS)
-        {
-          if (has_vert && has_horiz)
-            return META_FRAME_CONTROL_RESIZE_NW;
-          else if (has_vert)
-            return META_FRAME_CONTROL_RESIZE_N;
-          else
-            return META_FRAME_CONTROL_RESIZE_W;
-        }
-      else if (y < (fgeom.top_height + RESIZE_EXTENDS) &&
-               x >= (fgeom.width - RESIZE_EXTENDS))
-        {
-          if (has_vert && has_horiz)
-            return META_FRAME_CONTROL_RESIZE_NE;
-          else if (has_vert)
-            return META_FRAME_CONTROL_RESIZE_N;
-          else
-            return META_FRAME_CONTROL_RESIZE_E;
-        }
-      else if (y >= (fgeom.height - fgeom.bottom_height - RESIZE_EXTENDS))
-        {
-          if (has_vert)
-            return META_FRAME_CONTROL_RESIZE_S;
-        }
-      else if (y <= TOP_RESIZE_HEIGHT)
-        {
-          if (has_vert)
-            return META_FRAME_CONTROL_RESIZE_N;
-        }
-      else if (x <= fgeom.left_width)
-        {
-          if (has_horiz)
-            return META_FRAME_CONTROL_RESIZE_W;
-        }
-      else if (x >= (fgeom.width - fgeom.right_width))
-        {
-          if (has_horiz)
-            return META_FRAME_CONTROL_RESIZE_E;
-        }
+  if (y >= (fgeom.height - fgeom.bottom_height - RESIZE_EXTENDS) &&
+      x >= (fgeom.width - fgeom.right_width - RESIZE_EXTENDS))
+    {
+      if (has_vert && has_horiz)
+        return META_FRAME_CONTROL_RESIZE_SE;
+      else if (has_vert)
+        return META_FRAME_CONTROL_RESIZE_S;
+      else if (has_horiz)
+        return META_FRAME_CONTROL_RESIZE_E;
     }
-  
-  return META_FRAME_CONTROL_NONE;
+  else if (y >= (fgeom.height - fgeom.bottom_height - RESIZE_EXTENDS) &&
+           x <= (fgeom.left_width + RESIZE_EXTENDS))
+    {
+      if (has_vert && has_horiz)
+        return META_FRAME_CONTROL_RESIZE_SW;
+      else if (has_vert)
+        return META_FRAME_CONTROL_RESIZE_S;
+      else if (has_horiz)
+        return META_FRAME_CONTROL_RESIZE_W;
+    }
+  else if (y < (fgeom.top_height + RESIZE_EXTENDS) &&
+           x < RESIZE_EXTENDS)
+    {
+      if (has_vert && has_horiz)
+        return META_FRAME_CONTROL_RESIZE_NW;
+      else if (has_vert)
+        return META_FRAME_CONTROL_RESIZE_N;
+      else if (has_horiz)
+        return META_FRAME_CONTROL_RESIZE_W;
+    }
+  else if (y < (fgeom.top_height + RESIZE_EXTENDS) &&
+           x >= (fgeom.width - RESIZE_EXTENDS))
+    {
+      if (has_vert && has_horiz)
+        return META_FRAME_CONTROL_RESIZE_NE;
+      else if (has_vert)
+        return META_FRAME_CONTROL_RESIZE_N;
+      else if (has_horiz)
+        return META_FRAME_CONTROL_RESIZE_E;
+    }
+  else if (y >= (fgeom.height - fgeom.bottom_height - RESIZE_EXTENDS))
+    {
+      if (has_vert)
+        return META_FRAME_CONTROL_RESIZE_S;
+    }
+  else if (y <= TOP_RESIZE_HEIGHT)
+    {
+      if (has_vert)
+        return META_FRAME_CONTROL_RESIZE_N;
+      else if (has_horiz)
+        return META_FRAME_CONTROL_TITLE;
+    }
+  else if (x <= fgeom.left_width)
+    {
+      if (has_horiz)
+        return META_FRAME_CONTROL_RESIZE_W;
+    }
+  else if (x >= (fgeom.width - fgeom.right_width))
+    {
+      if (has_horiz)
+        return META_FRAME_CONTROL_RESIZE_E;
+    }
+
+  if (y >= bottom_of_titlebar)
+    return META_FRAME_CONTROL_NONE;
+  else
+    return META_FRAME_CONTROL_TITLE;
 }
 
 void
