@@ -79,6 +79,11 @@ static void handle_toggle_desktop     (MetaDisplay    *display,
                                        MetaWindow     *window,
                                        XEvent         *event,
                                        MetaKeyBinding *binding);
+static void handle_panel_keybinding   (MetaDisplay    *display,
+                                       MetaScreen     *screen,
+                                       MetaWindow     *window,
+                                       XEvent         *event,
+                                       MetaKeyBinding *binding);
 static void handle_toggle_maximize    (MetaDisplay    *display,
                                        MetaScreen     *screen,
                                        MetaWindow     *window,
@@ -209,7 +214,7 @@ struct _MetaKeyBinding
   const char *name;
   KeySym keysym;
   unsigned int mask;
-  int keycode;
+  unsigned int keycode;
   MetaVirtualModifier modifiers;
   const MetaKeyHandler *handler;
 };
@@ -265,6 +270,10 @@ static const MetaKeyHandler screen_handlers[] = {
     GINT_TO_POINTER (META_TAB_LIST_DOCKS) },  
   { META_KEYBINDING_SHOW_DESKTOP, handle_toggle_desktop,
     NULL },
+  { META_KEYBINDING_PANEL_MAIN_MENU, handle_panel_keybinding,
+    GINT_TO_POINTER (META_KEYBINDING_ACTION_PANEL_MAIN_MENU) },
+  { META_KEYBINDING_PANEL_RUN_DIALOG, handle_panel_keybinding,
+    GINT_TO_POINTER (META_KEYBINDING_ACTION_PANEL_RUN_DIALOG) },
   { META_KEYBINDING_COMMAND_1, handle_run_command,
     GINT_TO_POINTER (0) },
   { META_KEYBINDING_COMMAND_2, handle_run_command,
@@ -329,6 +338,10 @@ static const MetaKeyHandler screen_handlers[] = {
     GINT_TO_POINTER (30) },
   { META_KEYBINDING_COMMAND_32, handle_run_command,
     GINT_TO_POINTER (31) },
+  { META_KEYBINDING_COMMAND_SCREENSHOT, handle_run_command,
+    GINT_TO_POINTER (32) },
+  { META_KEYBINDING_COMMAND_WIN_SCREENSHOT, handle_run_command,
+    GINT_TO_POINTER (33) },
   { NULL, NULL, NULL }
 };
   
@@ -429,7 +442,7 @@ reload_modmap (MetaDisplay *display)
       /* get the key code at this point in the map,
        * see if its keysym is one we're interested in
        */
-      int keycode = modmap->modifiermap[i];
+      unsigned int keycode = modmap->modifiermap[i];
       
       if (keycode >= display->min_keycode &&
           keycode <= display->max_keycode)
@@ -512,7 +525,7 @@ reload_keycodes (MetaDisplay *display)
       int i;
       
       i = 0;
-      while (display->screen_bindings[i].keysym != None)
+      while (i < display->n_screen_bindings)
         {
           display->screen_bindings[i].keycode = XKeysymToKeycode (display->xdisplay,
                                                                   display->screen_bindings[i].keysym);
@@ -526,7 +539,7 @@ reload_keycodes (MetaDisplay *display)
       int i;
       
       i = 0;
-      while (display->window_bindings[i].keysym != None)
+      while (i < display->n_window_bindings)
         {
           display->window_bindings[i].keycode = XKeysymToKeycode (display->xdisplay,
                                                                   display->window_bindings[i].keysym);
@@ -547,11 +560,17 @@ reload_modifiers (MetaDisplay *display)
       int i;
       
       i = 0;
-      while (display->screen_bindings[i].keysym != None)
+      while (i < display->n_screen_bindings)
         {
           meta_display_devirtualize_modifiers (display,
                                                display->screen_bindings[i].modifiers,
                                                &display->screen_bindings[i].mask);
+
+          meta_topic (META_DEBUG_KEYBINDINGS,
+                      " Devirtualized mods 0x%x -> 0x%x (%s)\n",
+                      display->screen_bindings[i].modifiers,
+                      display->screen_bindings[i].mask,
+                      display->screen_bindings[i].name);          
           
           ++i;
         }
@@ -562,93 +581,137 @@ reload_modifiers (MetaDisplay *display)
       int i;
       
       i = 0;
-      while (display->window_bindings[i].keysym != None)
+      while (i < display->n_window_bindings)
         {
           meta_display_devirtualize_modifiers (display,
                                                display->window_bindings[i].modifiers,
                                                &display->window_bindings[i].mask);
+
+          meta_topic (META_DEBUG_KEYBINDINGS,
+                      " Devirtualized mods 0x%x -> 0x%x (%s)\n",
+                      display->window_bindings[i].modifiers,
+                      display->window_bindings[i].mask,
+                      display->window_bindings[i].name);
           
           ++i;
         }
     }
 }
 
+static int
+count_bindings (const MetaKeyPref *prefs,
+                int                n_prefs)
+{
+  int i;
+  int count;
+
+  count = 0;
+  i = 0;
+  while (i < n_prefs)
+    {
+      if (prefs[i].keysym != None)
+        {
+          count += 1;
+
+          if (prefs[i].add_shift &&
+              (prefs[i].modifiers & META_VIRTUAL_SHIFT_MASK) == 0)
+            count += 1;
+        }
+      
+      ++i;
+    }
+
+  return count;
+}
+
 static void
-rebuild_screen_binding_table (MetaDisplay *display)
+rebuild_binding_table (MetaDisplay        *display,
+                       MetaKeyBinding    **bindings_p,
+                       int                *n_bindings_p,
+                       const MetaKeyPref  *prefs,
+                       int                 n_prefs)
 {
-  const MetaKeyPref *prefs;
   int n_bindings;
   int src, dest;
   
-  meta_topic (META_DEBUG_KEYBINDINGS,
-              "Rebuilding screen binding table from preferences\n");
-  
-  meta_prefs_get_screen_bindings (&prefs, &n_bindings);
-  g_free (display->screen_bindings);
-  display->screen_bindings = g_new0 (MetaKeyBinding, n_bindings);
+  n_bindings = count_bindings (prefs, n_prefs);
+  g_free (*bindings_p);
+  *bindings_p = g_new0 (MetaKeyBinding, n_bindings);
 
   src = 0;
   dest = 0;
-  while (src < n_bindings)
+  while (src < n_prefs)
     {
       if (prefs[src].keysym != None)
         {
-          display->screen_bindings[dest].name = prefs[src].name;
-          display->screen_bindings[dest].keysym = prefs[src].keysym;
-          display->screen_bindings[dest].modifiers = prefs[src].modifiers;
-          display->screen_bindings[dest].mask = 0;
-          display->screen_bindings[dest].keycode = 0;          
+          (*bindings_p)[dest].name = prefs[src].name;
+          (*bindings_p)[dest].keysym = prefs[src].keysym;
+          (*bindings_p)[dest].modifiers = prefs[src].modifiers;
+          (*bindings_p)[dest].mask = 0;
+          (*bindings_p)[dest].keycode = 0;          
           
           ++dest;
+
+          if (prefs[src].add_shift &&
+              (prefs[src].modifiers & META_VIRTUAL_SHIFT_MASK) == 0)
+            {
+              meta_topic (META_DEBUG_KEYBINDINGS,
+                          "Binding %s also needs Shift grabbed\n",
+                          prefs[src].name);
+              
+              (*bindings_p)[dest].name = prefs[src].name;
+              (*bindings_p)[dest].keysym = prefs[src].keysym;
+              (*bindings_p)[dest].modifiers = prefs[src].modifiers |
+                META_VIRTUAL_SHIFT_MASK;
+              (*bindings_p)[dest].mask = 0;
+              (*bindings_p)[dest].keycode = 0;          
+              
+              ++dest;
+            }
         }
       
       ++src;
     }
 
-  display->n_screen_bindings = dest;
+  g_assert (dest == n_bindings);
+  
+  *n_bindings_p = dest;
 
   meta_topic (META_DEBUG_KEYBINDINGS,
-              "%d screen bindings in table\n",
-              display->n_screen_bindings);
+              " %d bindings in table\n",
+              *n_bindings_p);
+}
+
+static void
+rebuild_screen_binding_table (MetaDisplay *display)
+{
+  const MetaKeyPref *prefs;
+  int n_prefs;
+  
+  meta_topic (META_DEBUG_KEYBINDINGS,
+              "Rebuilding screen binding table from preferences\n");
+  
+  meta_prefs_get_screen_bindings (&prefs, &n_prefs);
+  rebuild_binding_table (display,
+                         &display->screen_bindings,
+                         &display->n_screen_bindings,
+                         prefs, n_prefs);
 }
 
 static void
 rebuild_window_binding_table (MetaDisplay *display)
 {
   const MetaKeyPref *prefs;
-  int n_bindings;
-  int src, dest;
+  int n_prefs;
   
   meta_topic (META_DEBUG_KEYBINDINGS,
               "Rebuilding window binding table from preferences\n");
   
-  meta_prefs_get_window_bindings (&prefs, &n_bindings);
-  g_free (display->window_bindings);
-  display->window_bindings = g_new0 (MetaKeyBinding, n_bindings);
-
-  src = 0;
-  dest = 0;
-  while (src < n_bindings)
-    {
-      if (prefs[src].keysym != None)
-        {
-          display->window_bindings[dest].name = prefs[src].name;
-          display->window_bindings[dest].keysym = prefs[src].keysym;
-          display->window_bindings[dest].modifiers = prefs[src].modifiers;
-          display->window_bindings[dest].mask = 0;
-          display->window_bindings[dest].keycode = 0;
-          
-          ++dest;
-        }
-      
-      ++src;
-    }
-
-  display->n_window_bindings = dest;
-
-  meta_topic (META_DEBUG_KEYBINDINGS,
-              "%d window bindings in table\n",
-              display->n_window_bindings);
+  meta_prefs_get_window_bindings (&prefs, &n_prefs);
+  rebuild_binding_table (display,
+                         &display->window_bindings,
+                         &display->n_window_bindings,
+                         prefs, n_prefs);
 }
 
 static void
@@ -854,10 +917,10 @@ meta_change_keygrab (MetaDisplay *display,
                      Window       xwindow,
                      gboolean     grab,
                      int          keysym,
-                     int          keycode,
+                     unsigned int keycode,
                      int          modmask)
 {
-  unsigned int ignored_mask;
+  int ignored_mask;
 
   /* Grab keycode/modmask, together with
    * all combinations of ignored modifiers.
@@ -874,7 +937,7 @@ meta_change_keygrab (MetaDisplay *display,
   meta_error_trap_push (display);
   
   ignored_mask = 0;
-  while (ignored_mask <= display->ignored_modifier_mask)
+  while (ignored_mask < (int) display->ignored_modifier_mask)
     {
       if (ignored_mask & ~(display->ignored_modifier_mask))
         {
@@ -925,7 +988,7 @@ static void
 meta_grab_key (MetaDisplay *display,
                Window       xwindow,
                int          keysym,
-               int          keycode,
+               unsigned int keycode,
                int          modmask)
 {
   meta_change_keygrab (display, xwindow, TRUE, keysym, keycode, modmask);
@@ -1017,6 +1080,14 @@ meta_window_grab_keys (MetaWindow  *window)
 {
   if (window->all_keys_grabbed)
     return;
+
+  if (window->type == META_WINDOW_DOCK)
+    {
+      if (window->keys_grabbed)
+        ungrab_all_keys (window->display, window->xwindow);
+      window->keys_grabbed = FALSE;
+      return;
+    }
   
   if (window->keys_grabbed)
     {
@@ -1422,6 +1493,11 @@ process_event (MetaKeyBinding       *bindings,
         {
           const MetaKeyHandler *handler;
 
+          meta_topic (META_DEBUG_KEYBINDINGS,
+                      "Binding keycode 0x%x mask 0x%x matches event 0x%x state 0x%x\n",
+                      bindings[i].keycode, bindings[i].mask,
+                      event->xkey.keycode, event->xkey.state);
+          
           if (bindings[i].handler)
             handler = bindings[i].handler;
           else
@@ -1618,14 +1694,26 @@ process_keyboard_move_grab (MetaDisplay *display,
   else
     incr = NORMAL_INCREMENT;
 
+  if (keysym == XK_Escape)
+    {
+      /* End move and restore to original position */
+      meta_window_move_resize (display->grab_window,
+                               TRUE,
+                               display->grab_initial_window_pos.x,
+                               display->grab_initial_window_pos.y,
+                               display->grab_initial_window_pos.width,
+                               display->grab_initial_window_pos.height);
+    }
+  
   /* When moving by increments, we still snap to edges if the move
    * to the edge is smaller than the increment. This is because
    * Shift + arrow to snap is sort of a hidden feature. This way
    * people using just arrows shouldn't get too frustrated.
    */
-      
   switch (keysym)
     {
+    case XK_KP_Home:
+    case XK_KP_Prior:
     case XK_Up:
     case XK_KP_Up:
       edge = meta_window_find_next_horizontal_edge (window, FALSE);
@@ -1636,6 +1724,8 @@ process_keyboard_move_grab (MetaDisplay *display,
           
       handled = TRUE;
       break;
+    case XK_KP_End:
+    case XK_KP_Next:
     case XK_Down:
     case XK_KP_Down:
       edge = meta_window_find_next_horizontal_edge (window, TRUE);
@@ -1646,6 +1736,12 @@ process_keyboard_move_grab (MetaDisplay *display,
           
       handled = TRUE;
       break;
+    }
+  
+  switch (keysym)
+    {
+    case XK_KP_Home:
+    case XK_KP_End:
     case XK_Left:
     case XK_KP_Left:
       edge = meta_window_find_next_vertical_edge (window, FALSE);
@@ -1656,6 +1752,8 @@ process_keyboard_move_grab (MetaDisplay *display,
 
       handled = TRUE;
       break;
+    case XK_KP_Prior:
+    case XK_KP_Next:
     case XK_Right:
     case XK_KP_Right:
       edge = meta_window_find_next_vertical_edge (window, TRUE);
@@ -1664,19 +1762,6 @@ process_keyboard_move_grab (MetaDisplay *display,
         x = edge;
       handled = TRUE;
       break;
-
-    case XK_Escape:
-      /* End move and restore to original position */
-      meta_window_move_resize (display->grab_window,
-                               TRUE,
-                               display->grab_initial_window_pos.x,
-                               display->grab_initial_window_pos.y,
-                               display->grab_initial_window_pos.width,
-                               display->grab_initial_window_pos.height);
-      break;
-          
-    default:
-      break;
     }
 
   if (handled)
@@ -2516,6 +2601,56 @@ handle_toggle_desktop (MetaDisplay    *display,
     meta_screen_show_desktop (screen);
 }
 
+static void
+handle_panel_keybinding (MetaDisplay    *display,
+                         MetaScreen     *screen,
+                         MetaWindow     *window,
+                         XEvent         *event,
+                         MetaKeyBinding *binding)
+{
+  MetaKeyBindingAction action;
+  Atom action_atom;
+  XClientMessageEvent ev;
+  
+  action = GPOINTER_TO_INT (binding->handler->data);
+
+  action_atom = None;
+  switch (action)
+    {
+    case META_KEYBINDING_ACTION_PANEL_MAIN_MENU:
+      action_atom = display->atom_gnome_panel_action_main_menu;
+      break;
+    case META_KEYBINDING_ACTION_PANEL_RUN_DIALOG:
+      action_atom = display->atom_gnome_panel_action_run_dialog;
+      break;
+    default:
+      return;
+    }
+   
+  ev.type = ClientMessage;
+  ev.window = screen->xroot;
+  ev.message_type = display->atom_gnome_panel_action;
+  ev.format = 32;
+  ev.data.l[0] = action_atom;
+  ev.data.l[1] = event->xkey.time;
+
+  meta_topic (META_DEBUG_KEYBINDINGS,
+              "Sending panel message with timestamp %lu\n", event->xkey.time);
+
+  meta_error_trap_push (display);
+
+  /* Release the grab for the panel before sending the event */
+  XUngrabKeyboard (display->xdisplay, event->xkey.time);
+
+  XSendEvent (display->xdisplay,
+	      screen->xroot,
+	      False,
+	      StructureNotifyMask,
+	      (XEvent*) &ev);
+
+  meta_error_trap_pop (display, FALSE);
+}
+
 static void
 handle_activate_menu (MetaDisplay    *display,
                       MetaScreen     *screen,
